from pathlib import Path

import pandas as pd
import pytest
from pandas.testing import assert_frame_equal

from wind_up.models import WindUpConfig
from wind_up.pp_analysis import pre_post_pp_analysis_with_reversal


def test_pre_post_pp_analysis_with_reversal(test_lsa_t13_config: WindUpConfig) -> None:
    cfg = test_lsa_t13_config

    test_wtg = next(x for x in cfg.asset.wtgs if x.name == "LSA_T13")
    ref_name = "LSA_T12"
    detrend_ws_col = "ref_ws_detrended"
    test_pw_col = "test_pw_clipped"
    ref_wd_col = "ref_YawAngleMean"

    pre_df = pd.read_parquet(Path(__file__).parents[0] / "test_data/LSA_T13_LSA_T12_pre_df.parquet")
    post_df = pd.read_parquet(Path(__file__).parents[0] / "test_data/LSA_T13_LSA_T12_post_df.parquet")
    lt_wtg_df_filt = pd.read_parquet(Path(__file__).parents[0] / "test_data/LSA_T13_lt_wtg_df_filt.parquet")
    test_df = pd.read_parquet(Path(__file__).parents[0] / "test_data/LSA_T13_test_df.parquet")
    expected_df = pd.read_parquet(Path(__file__).parents[0] / "test_data/pre_post_pp_analysis_expected_df.parquet")
    pp_results, actual_df = pre_post_pp_analysis_with_reversal(
        cfg=cfg,
        test_wtg=test_wtg,
        ref_name=ref_name,
        lt_df=lt_wtg_df_filt,
        pre_df=pre_df,
        post_df=post_df,
        ws_col=detrend_ws_col,
        pw_col=test_pw_col,
        wd_col=ref_wd_col,
        plot_cfg=None,
        test_df=test_df,
    )

    # minor changes to make actual_df compatible with old expected_df
    expected_df["hours_for_mwh_calc"] = expected_df["hours_per_year"]
    expected_df["hours_per_year"] = actual_df["hours_per_year"]
    cols_with_new_calc = ["uplift_kw_se", "uplift_p5_kw", "uplift_p95_kw"]
    expected_df[cols_with_new_calc] = actual_df[cols_with_new_calc]
    new_cols = [
        "pre_valid",
        "post_valid",
        "hours_pre_raw",
        "hours_post_raw",
        "is_invalid_bin",
        "pw_at_mid_expected",
        "pw_sem_at_mid_expected",
        "relative_cp_baseline",
        "relative_cp_post",
        "relative_cp_sem_at_mid_expected",
        "relative_cp_sem_at_mid_post",
        "uplift_relative_cp",
        "uplift_relative_cp_se",
        "uplift_relative_cp_p5",
        "uplift_relative_cp_p95",
    ]
    expected_df[new_cols] = actual_df[new_cols]
    expected_df = expected_df[actual_df.columns]

    assert_frame_equal(actual_df, expected_df)
    assert pp_results["pp_valid_hours"] == pytest.approx(10748.5)
    assert pp_results["pp_valid_hours_pre"] == pytest.approx(5807.333333333333)
    assert pp_results["pp_valid_hours_post"] == pytest.approx(4941.166666666667)
    assert pp_results["pp_invalid_bin_count"] == 3
    assert pp_results["pp_data_coverage"] == pytest.approx(0.6793388952092024)
    assert pp_results["reversal_error"] == pytest.approx(-0.008786551768533796)
    assert pp_results["uplift_noadj_frc"] == pytest.approx(0.04523448345231426)
    assert pp_results["poweronly_uplift_frc"] == pytest.approx(0.04560411838169785)
    assert pp_results["reversed_uplift_frc"] == pytest.approx(0.03681756661316406)
    assert pp_results["uplift_frc"] == pytest.approx(0.040841207568047364)
    assert pp_results["missing_bins_unc_scale_factor"] == pytest.approx(1.0000000006930523)
    assert pp_results["t_value_one_sigma"] == pytest.approx(1.0000168636907854)
    assert pp_results["unc_one_sigma_lowerbound_frc"] == pytest.approx(0.004393275884266898)
    assert pp_results["unc_one_sigma_frc"] == pytest.approx(0.004393275884266898)



import matplotlib.pyplot as plt
import pandas as pd

from wind_up.constants import RAW_POWER_COL, RAW_WINDSPEED_COL, SCATTER_ALPHA, SCATTER_MARKERSCALE, SCATTER_S
from wind_up.models import PlotConfig


def plot_waking_state_one_ttype_or_wtg(wf_df: pd.DataFrame, ttype_or_wtg: str, plot_cfg: PlotConfig) -> None:
    waking_df = wf_df[wf_df["waking"]]
    not_waking_df = wf_df[wf_df["not_waking"]]
    unknown_df = wf_df[wf_df["unknown_waking"]]
    plt.figure()
    plt.scatter(
        waking_df[RAW_WINDSPEED_COL],
        waking_df[RAW_POWER_COL],
        label="waking",
        s=SCATTER_S,
        alpha=SCATTER_ALPHA,
    )
    plt.scatter(
        not_waking_df[RAW_WINDSPEED_COL],
        not_waking_df[RAW_POWER_COL],
        label="not waking",
        s=SCATTER_S,
        alpha=SCATTER_ALPHA,
    )
    plt.scatter(
        unknown_df[RAW_WINDSPEED_COL],
        unknown_df[RAW_POWER_COL],
        label="unknown or partial waking",
        s=SCATTER_S,
        alpha=SCATTER_ALPHA,
    )
    plt.xlabel(f"{RAW_WINDSPEED_COL} [m/s]")
    plt.ylabel(f"{RAW_POWER_COL} [kW]")
    plot_title = f"{ttype_or_wtg} power curve by waking state"
    plt.title(plot_title)
    plt.grid()
    plt.legend(loc="lower right", markerscale=SCATTER_MARKERSCALE)
    plt.tight_layout()
    if plot_cfg.show_plots:
        plt.show()
    if plot_cfg.save_plots:
        plt.savefig(plot_cfg.plots_dir / ttype_or_wtg / f"{plot_title}.png")
    plt.close()



import pandas as pd
from matplotlib import pyplot as plt

from wind_up.constants import RAW_POWER_COL, REANALYSIS_WD_COL, WINDFARM_YAWDIR_COL
from wind_up.models import PlotConfig, WindUpConfig


def plot_diff_to_north_ref_wd(
    wtg_df: pd.DataFrame,
    *,
    wtg_name: str,
    north_ref_wd_col: str,
    loop_count: int,
    plot_cfg: PlotConfig,
) -> None:
    plt.figure()
    plt.plot(wtg_df.index, wtg_df[f"yaw_diff_to_{north_ref_wd_col}"], label=f"yaw_diff_to_{north_ref_wd_col}")
    plt.plot(wtg_df.index, wtg_df[f"filt_diff_to_{north_ref_wd_col}"], label=f"filt_diff_to_{north_ref_wd_col}")
    plt.plot(
        wtg_df.index,
        wtg_df[f"short_rolling_diff_to_{north_ref_wd_col}"],
        label=f"short_rolling_diff_to_{north_ref_wd_col}",
    )
    plt.plot(
        wtg_df.index,
        wtg_df[f"long_rolling_diff_to_{north_ref_wd_col}"],
        label=f"long_rolling_diff_to_{north_ref_wd_col}",
    )
    plt.grid()
    plt.legend()
    title = f"{wtg_name} diff to {north_ref_wd_col} loop_count={loop_count}"
    plt.title(title)
    plt.xlabel("datetime")
    plt.ylabel(f"yaw angle diff to {north_ref_wd_col} [deg]")
    plt.tight_layout()
    plt.savefig(plot_cfg.plots_dir / wtg_name / f"{title}.png")
    plt.close()


def plot_yaw_diff_vs_power(wtg_df: pd.DataFrame, *, wtg_name: str, north_ref_wd_col: str, plot_cfg: PlotConfig) -> None:
    plt.figure()
    plt.scatter(wtg_df[RAW_POWER_COL], wtg_df[f"yaw_diff_to_{north_ref_wd_col}"], s=1, alpha=0.2)
    plt.grid()
    plt.xlabel(RAW_POWER_COL)
    plt.ylabel(f"yaw_diff_to_{north_ref_wd_col}")
    title = f"{wtg_name} yaw_diff_to_{north_ref_wd_col} vs {RAW_POWER_COL}"
    plt.tight_layout()
    (plot_cfg.plots_dir / wtg_name).mkdir(exist_ok=True)
    plt.savefig(plot_cfg.plots_dir / wtg_name / f"{title}.png")
    plt.close()


def plot_wf_yawdir_and_reanalysis_timeseries(wf_df: pd.DataFrame, *, cfg: WindUpConfig, plot_cfg: PlotConfig) -> None:
    days_to_plot = 60
    plot_df = wf_df.loc[cfg.asset.wtgs[0].name].copy()
    last_reanalysis_dt = plot_df[REANALYSIS_WD_COL].dropna().index.max()
    plot_start_dt = last_reanalysis_dt - pd.Timedelta(days=days_to_plot)
    plot_df = plot_df[plot_start_dt:]

    plt.figure()
    plt.plot(plot_df.index, plot_df[WINDFARM_YAWDIR_COL], label=WINDFARM_YAWDIR_COL, marker=".", linestyle="None")
    plt.plot(plot_df.index, plot_df[REANALYSIS_WD_COL], label=REANALYSIS_WD_COL)
    plt.xlabel("timestamp")
    plt.ylabel("direction [deg]")
    title = "wf_yawdir and reanalysis timeseries"
    plt.title(title)
    plt.legend()
    plt.grid()
    plt.tight_layout()
    plt.savefig(plot_cfg.plots_dir / f"{title}.png")
    plt.close()



import logging

import numpy as np
import pandas as pd

from wind_up.constants import DEFAULT_AIR_DENSITY
from wind_up.models import PlotConfig, TurbineType, WindUpConfig
from wind_up.plots.ws_est_plots import plot_ws_est_gain_xs_one_ttype, plot_ws_est_one_ttype_or_wtg
from wind_up.wind_funcs import calc_cp

logger = logging.getLogger(__name__)


def calc_pc_low_high_one_ttype(
    df: pd.DataFrame,
    x_col: str,
    y_col: str,
    x_bin_width: float,
    low_q_pct: float,
    high_q_pct: float,
) -> pd.DataFrame:
    x_bin_edges = np.arange(0, df[x_col].max() + x_bin_width, x_bin_width)
    return df.groupby(by=pd.cut(df[x_col], bins=x_bin_edges, retbins=False), observed=True).agg(
        x_mean=pd.NamedAgg(column=x_col, aggfunc=lambda x: x.mean()),
        y_low=pd.NamedAgg(column=y_col, aggfunc=lambda x: np.nanpercentile(x, low_q_pct)),
        y_high=pd.NamedAgg(column=y_col, aggfunc=lambda x: np.nanpercentile(x, high_q_pct)),
    )


def add_ws_est_one_ttype(
    cfg: WindUpConfig,
    df: pd.DataFrame,
    ttype: TurbineType,
    pc: pd.DataFrame,
    plot_cfg: PlotConfig | None,
) -> pd.DataFrame:
    df_input = df.copy()

    # at some point would be good to use an air density time series
    df_input["cp"] = calc_cp(
        power_kw=df_input["pw_clipped"],
        ws_ms=df_input["WindSpeedMean"],
        air_density_kgpm3=DEFAULT_AIR_DENSITY,
        rotor_diameter_m=ttype.rotor_diameter_m,
    )
    df_input["cp"] = df_input["cp"].clip(upper=2)
    ws_half_rated = np.interp(ttype.rated_power_kw / 2, pc["pw_clipped"].values, pc["WindSpeedMean"].values)
    cp_calc_ws_range = 1
    target_cp = df_input["cp"][(df_input["WindSpeedMean"] - ws_half_rated).abs() < cp_calc_ws_range / 2].mean()

    for wtg in df_input.index.unique(level="TurbineName"):
        df_wtg = df_input.loc[wtg]
        mean_cp_wtg = df_wtg.loc[(df_wtg["WindSpeedMean"] - ws_half_rated).abs() < cp_calc_ws_range / 2, "cp"].mean()
        cp_correction_factor = mean_cp_wtg ** (1 / 3) / target_cp ** (1 / 3)
        logger.info(f"{wtg} cp correction factor = {cp_correction_factor:.2f}")
        df_input.loc[wtg, "ws_cp_corrected"] = cp_correction_factor * df_input.loc[wtg, "WindSpeedMean"].to_numpy()

    # find the four wind speeds used for gain
    low_q_pct = 0.01
    pc_low_high = calc_pc_low_high_one_ttype(
        df=df_input,
        x_col="ws_cp_corrected",
        y_col="pw_clipped",
        x_bin_width=cfg.ws_bin_width / 2,
        low_q_pct=low_q_pct,
        high_q_pct=100 - low_q_pct,
    )
    if pc_low_high["y_high"].min() < (ttype.rated_power_kw * 0.01):
        ws0 = float(np.interp(ttype.rated_power_kw * 0.01, pc_low_high["y_high"], pc_low_high["x_mean"]))
    else:
        ws0 = 0
    ws1 = float(np.interp(ttype.rated_power_kw * 0.01, pc_low_high["y_low"], pc_low_high["x_mean"]))
    ws1 = max(ws0 + 1, ws1)
    high_power_ws = min(ttype.cutout_ws_mps - 3, 17)
    high_power_threshold = pc_low_high["y_low"][pc_low_high["x_mean"] >= high_power_ws].min() * 0.99
    ws2 = float(np.interp(high_power_threshold, pc_low_high["y_high"], pc_low_high["x_mean"]))
    ws3 = float(np.interp(high_power_threshold, pc_low_high["y_low"], pc_low_high["x_mean"]))
    ws3 = max(ws2 + 1, ws3)
    if plot_cfg is not None:
        plot_ws_est_gain_xs_one_ttype(
            pc_low_high=pc_low_high,
            ttype=ttype.turbine_type,
            rated_power_kw=ttype.rated_power_kw,
            x0=ws0,
            x1=ws1,
            x2=ws2,
            x3=ws3,
            plot_cfg=plot_cfg,
        )
    # gain 1 uses wind speed as x axis
    ws_est_gain1_x = [ws0, ws1, ws2, ws3]
    ws_est_gain1_y = [0, 1, 1, -1]
    if not np.all(np.diff(ws_est_gain1_x) > 0):
        msg = "x values for gain 1 must be increasing"
        raise RuntimeError(msg)
    ws_est_gain1 = np.interp(df_input["ws_cp_corrected"].values, ws_est_gain1_x, ws_est_gain1_y)

    # gain 2 uses power as x axis
    ws_est_gain2_x = [0, 0.1 * ttype.rated_power_kw, 0.9 * ttype.rated_power_kw, ttype.rated_power_kw]
    ws_est_gain2_y = [0, 1, 1, 0]
    if not np.all(np.diff(ws_est_gain2_x) > 0):
        msg = "x values for gain 2 must be increasing"
        raise RuntimeError(msg)
    ws_est_gain2 = np.interp(df_input["pw_clipped"].values, ws_est_gain2_x, ws_est_gain2_y)

    # combine the two gains as a simple average
    df_input["ws_est_gain"] = (ws_est_gain1 + ws_est_gain2) / 2
    df_input["ws_est_gain"] = df_input["ws_est_gain"].clip(lower=0, upper=1)

    pc_transposed = df_input.groupby(
        by=pd.qcut(df_input["pw_clipped"], q=50, retbins=False, duplicates="drop"),
        observed=True,
    ).agg(
        p_bin=pd.NamedAgg(column="pw_clipped", aggfunc=lambda x: x.mean()),
        ws_cp_corrected=pd.NamedAgg(column="ws_cp_corrected", aggfunc=lambda x: x.mean()),
    )
    pc_transposed = pc_transposed.set_index("p_bin")

    df_input["ws_est_from_power_only"] = np.interp(
        df_input["pw_clipped"].values,
        pc_transposed.index.values,
        pc_transposed["ws_cp_corrected"].values,
    )
    df_input["ws_est_blend"] = (
        df_input["ws_est_gain"] * df_input["ws_est_from_power_only"]
        + (1 - df_input["ws_est_gain"]) * df_input["ws_cp_corrected"]
    )
    if plot_cfg is not None:
        plot_ws_est_one_ttype_or_wtg(
            df=df_input,
            ttype_or_wtg=ttype.turbine_type,
            pc_transposed=pc_transposed,
            plot_cfg=plot_cfg,
        )
        if not plot_cfg.skip_per_turbine_plots:
            for wtg in df_input.index.unique(level="TurbineName"):
                df_wtg = df_input.loc[wtg]
                plot_ws_est_one_ttype_or_wtg(
                    df=df_wtg, ttype_or_wtg=wtg, pc_transposed=pc_transposed, plot_cfg=plot_cfg
                )

    cols_to_return = ["ws_est_from_power_only", "ws_est_blend"]
    df[cols_to_return] = df_input[cols_to_return]
    return df


def add_ws_est(cfg: WindUpConfig, wf_df: pd.DataFrame, pc_per_ttype: dict, plot_cfg: PlotConfig | None) -> pd.DataFrame:
    _msg = "#" * 78 + "\n# estimate wind speed from power\n" + "#" * 78
    logger.info(_msg)
    df_input = wf_df.copy()
    wf_df = pd.DataFrame()
    for ttype in cfg.list_unique_turbine_types():
        wtgs = cfg.list_turbine_ids_of_type(ttype)
        df_ttype = df_input.loc[wtgs]
        df_ = add_ws_est_one_ttype(
            cfg=cfg,
            df=df_ttype,
            ttype=ttype,
            pc=pc_per_ttype[ttype.turbine_type],
            plot_cfg=plot_cfg,
        )
        wf_df = pd.concat([wf_df, df_])
    return wf_df.sort_index()



import pandas as pd

from wind_up.models import PrePost, WindUpConfig


def test_lsa_asset_name(test_lsa_t13_config: WindUpConfig) -> None:
    assert test_lsa_t13_config.asset.name == "Lisa Wind Farm"


def test_brt_asset_name(test_brt_t16_pitch: WindUpConfig) -> None:
    assert test_brt_t16_pitch.asset.name == "Bart Wind Farm"


def test_datetimes_lsa_t13(test_lsa_t13_config: WindUpConfig) -> None:
    cfg = test_lsa_t13_config
    assert isinstance(cfg.prepost, PrePost)
    assert cfg.prepost.post_first_dt_utc_start == pd.Timestamp("2021-09-30 00:00:00", tz="UTC")
    assert cfg.prepost.post_last_dt_utc_start == pd.Timestamp("2022-07-20 23:50:00", tz="UTC")
    assert cfg.prepost.pre_first_dt_utc_start == pd.Timestamp("2020-09-29 18:00:00", tz="UTC")
    assert cfg.prepost.pre_last_dt_utc_start == pd.Timestamp("2021-07-20 17:50:00", tz="UTC")
    assert cfg.analysis_first_dt_utc_start == pd.Timestamp("2020-09-29 18:00:00", tz="UTC")
    assert cfg.analysis_last_dt_utc_start == pd.Timestamp("2022-07-20 23:50:00", tz="UTC")
    assert cfg.lt_first_dt_utc_start == pd.Timestamp("2018-07-20 23:50:00", tz="UTC")
    assert cfg.lt_last_dt_utc_start == pd.Timestamp("2021-07-20 17:50:00", tz="UTC")


def test_datetimes_brt_t16(test_brt_t16_pitch: WindUpConfig) -> None:
    cfg = test_brt_t16_pitch
    assert cfg.analysis_first_dt_utc_start == pd.Timestamp("2022-11-03 15:30:00", tz="UTC")
    assert cfg.analysis_last_dt_utc_start == pd.Timestamp("2023-06-08 23:50:00", tz="UTC")
    assert cfg.lt_first_dt_utc_start == pd.Timestamp("2019-10-27 21:30:00", tz="UTC")
    assert cfg.lt_last_dt_utc_start == pd.Timestamp("2022-10-27 15:30:00", tz="UTC")


def test_get_max_rated_power_marge(test_marge_config: WindUpConfig) -> None:
    cfg = test_marge_config
    assert cfg.get_max_rated_power() == 1900


def test_get_max_rated_power_homer(test_homer_config: WindUpConfig) -> None:
    # Homer Wind Farm has two rated powers, 1000 and 1330
    cfg = test_homer_config
    assert cfg.get_max_rated_power() == 1330


def test_list_unique_turbine_types_marge(test_marge_config: WindUpConfig) -> None:
    cfg = test_marge_config
    assert isinstance(cfg.list_unique_turbine_types(), list)
    assert len(cfg.list_unique_turbine_types()) == 1
    assert cfg.list_unique_turbine_types()[0].turbine_type == "A90-1.9MW-90"


def test_list_unique_turbine_types_homer(test_homer_config: WindUpConfig) -> None:
    cfg = test_homer_config
    assert len(cfg.list_unique_turbine_types()) == 2
    assert cfg.list_unique_turbine_types()[0].turbine_type == "B-1.0-62"
    assert cfg.list_unique_turbine_types()[1].turbine_type == "C-1.3-62"


def test_list_turbine_ids_of_type_marge(test_marge_config: WindUpConfig) -> None:
    cfg = test_marge_config
    tt_list = cfg.list_unique_turbine_types()
    assert len(cfg.list_turbine_ids_of_type(tt_list[0])) == 9
    assert cfg.list_turbine_ids_of_type(tt_list[0])[0] == "MRG_T01"
    assert cfg.list_turbine_ids_of_type(tt_list[0])[-1] == "MRG_T09"


def test_list_turbine_ids_of_type_homer(test_homer_config: WindUpConfig) -> None:
    cfg = test_homer_config
    tt_list = cfg.list_unique_turbine_types()
    assert len(cfg.list_turbine_ids_of_type(tt_list[0])) == 1
    assert len(cfg.list_turbine_ids_of_type(tt_list[1])) == 1
    assert cfg.list_turbine_ids_of_type(tt_list[0])[0] == "HMR_T01"
    assert cfg.list_turbine_ids_of_type(tt_list[1])[0] == "HMR_T02"


def test_get_normal_operation_genrpm_range_marge(test_marge_config: WindUpConfig) -> None:
    cfg = test_marge_config
    tt_list = cfg.list_unique_turbine_types()
    assert cfg.get_normal_operation_genrpm_range(tt_list[0]) == (800, 1600)


def test_get_normal_operation_genrpm_range_homer(test_homer_config: WindUpConfig) -> None:
    cfg = test_homer_config
    tt_list = cfg.list_unique_turbine_types()
    assert cfg.get_normal_operation_genrpm_range(tt_list[0]) == (950, 1550)
    assert cfg.get_normal_operation_genrpm_range(tt_list[1]) == (950, 1550)


def test_get_normal_operation_pitch_range_marge(test_marge_config: WindUpConfig) -> None:
    cfg = test_marge_config
    tt_list = cfg.list_unique_turbine_types()
    assert cfg.get_normal_operation_pitch_range(tt_list[0]) == (-10, 40)


def test_get_normal_operation_pitch_range_homer(test_homer_config: WindUpConfig) -> None:
    cfg = test_homer_config
    tt_list = cfg.list_unique_turbine_types()
    assert cfg.get_normal_operation_pitch_range(tt_list[0]) == (-10, 5)
    assert cfg.get_normal_operation_pitch_range(tt_list[1]) == (-10, 5)


def test_turbine_get_latlongs(test_homer_config: WindUpConfig) -> None:
    cfg = test_homer_config
    assert cfg.asset.wtgs[0].get_latlongs() == [(-58.60364145072843, 103.6841410289202052)]
    assert cfg.asset.wtgs[1].get_latlongs() == [(-58.60261454305449, 103.688364968451364)]
